#  NetTree.pm - a package to encapusate the subnet allocations within a network
#  Copyright (C) 1998  Dave Plonka
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

# $Id: NetTree.pm,v 1.20 1999/07/29 00:20:19 dplonka Exp $
# Dave Plonka <plonka@doit.wisc.edu>

use strict;

package NetTree;

require 5;
require Exporter;

@NetTree::ISA=qw(Exporter);
@NetTree::EXPORT=qw(octets2bits bits2octets);
# convert the RCS revision to a reasonable Exporter VERSION:
'$Revision: 1.20 $' =~ m/(\d+)\.(\d+)/ && (( $NetTree::VERSION ) = sprintf("%d.%03d", $1, $2));

=head1 NAME

NetTree - a package to encapusate the subnet allocations within a network

=head1 SYNOPSIS

   use NetTree;

   $tree = new NetTree '10.0.0.0/8';
   %collisions = $tree->add('10.0.0.0/24', 'my subnet');
   %collisions = $tree->add('10.0.1.0/24', 'my subnet too');
   $tree->prune;
   @available = $tree->available;
   grep { print "$_\n" } @available;
   %unavailable = $tree->unavailable;
   while (($subnet, $cargo) = each %unavailable) {
       print "$subnet $cargo\n"
   }

   printf "%s\n", octets2bits('10.0.0.0/8');
   printf "%s\n", scalar(bits2octets(octets2bits('10.0.0.0/8')));

=head1 DESCRIPTION

B<NetTree> is a package used to define, and perform operations upon, networks.
A B<NetTree> object encapsulates a binary search tree of nodes that define
the subnets within that network.

B<NetTree.pm> was originally conceived of, and written to validate
subnet specifications which, in the past, were stored using simpler
methods such as a flat file with one subnet (and its description) per
line.  With such a simple data structure (essentially an array of subnet
specifications) it was difficult to enforce that subnet specifications
do not define overlapping networks (due to typos or administrative
accidents).  It was similarly difficult to find which subnet contains
a given (smaller) subnet or host.  The binary search tree representation
solved these problems and is a fairly frugal representation of even a
sparsely populated network.

The following methods and subroutines are defined:

=over 4

=cut

# { These are array indexes into the arrays in the binary search tree:
# (Elements 0 and 1 are "pointers" to other nodes in the tree.)
$NetTree::NET = 2;
$NetTree::SPEC = 3;
$NetTree::CARGO = 4;
# }

=item new

The B<new> method constructs and returns a B<NetTree> object.  It has one
optional argument specifying the network number and mask width.  If not
supplied, it essentially defaults to "0.0.0.0/0" - that is, the network
containing all 32-bit IP addresses.  Note that, if possible, it is
advantageous for performance reasons to specify the network and mask width.
However, this can only be done if one knows the network in which all
subsequently added subnets will reside, for instance the class A, B or C
network that your organization uses.

=cut

sub new {
   my $class = shift;
   my $netspec = shift;
   my $self = {};
   return undef unless _init($self, $netspec);
   return bless $self, $class
}

sub _init {
   my $self = shift;
   return undef unless ref($self);
   my $netspec = shift;
   $self->{bits} = octets2bits($netspec);
   $self->{tree} = [];
   $self->{cache} = {};
   return $self
}

=item octets2bits

The B<NetTree::octets2bits> subroutine takes a network specification as
either one scalar argument in "n.n.n.n/n" format, or an array of the
individual digits and mask width.  If the network mask width is not
specified, it is assumed to be 32 - that is an individual host address.
This subroutine returns a network specification as a bit string.  This
bit string is of the mask width characters in length.

=cut

# convert a "dotted.decimal/netmaskwidth" specification to a bit string
sub octets2bits {
   my @octet;
   my $netmasklen;
   if (0 == $#_) {
      my $net = shift;
      if ($net !~ m:^(\d+\.\d+\.\d+\.\d+)(/(\d+))?:) {
         return undef
      }
      @octet = split(m/\./, $1);
      $netmasklen = $3? $3 : 8*(1+$#octet)
   } else {
      $netmasklen = pop(@_);
      @octet = @_
   }
   return unpack("B${netmasklen}", pack("C*", @octet))
}

=item bits2octets

The B<NetTree::octets2bits> subroutine takes a network specification in
bit string format (such as that returned by B<NetTree::octets2bits>) as
an argument.
In a scalar context, this subroutine returns a network specification in
"n.n.n.n/n" format.  In an array context an array of the individual
digits and the network mask width is returned.

=cut

# convert a bit string to a "dotted.decimal/netmaskwidth" specification
# returns:
#    in a scalar context, a "dotted.decimal/netmaskwidth" specification
#    in an array context, an array of the digits followed by the net mask width
sub bits2octets {
   my $bits = shift; # a string of bits (should be a multiple of 8 in length)
   my $len = length $bits;
   my $ip = join('.', unpack("C*", pack("B*", $bits)), split('', '0' x (4-($len/8))));

   if (wantarray) {
       (split(m/\./, $ip), $len)
   } else {
       $ip . '/' . $len
   }
}

=item subnet

The B<subnet> method takes a network or host specification in "n.n.n.n/n"
format.  If the network mask width is unspecified it is assumed to be 32 -
that is, an individual host address specification.  B<subnet> returns a
hash of subnets in which the specified network or host resides.  (Usually
only one element would be returned in the hash.)

The hash's key is in "n.n.n.n/n" format and the corresponding value is
whatever node value was specified when that subnet was B<add>ed to the tree.

=cut

# find the subnet to which the specified subnet or host belongs
# e.g.
#
#    %subnet = $this->subnet('128.104.31.31')
#
# returns:
# on success, a hash of subnet(s).
#    The key to this hash is the subnet specification, and the value is the
#    "cargo" for that subnet.
# on failure, an empty list.
#
sub subnet {
    my $self = shift;
    my $spec = shift;
    if (defined($self->{cache}->{$spec})) {
       return @{$self->{cache}->{$spec}}
    }
    my @subnet = $self->add($spec);
    die if (1 < $#subnet); # should only yield, at most, one subnet
    if (@subnet) {
       $self->{cache}->{$spec} = [@subnet]
    }
    return @subnet
}

=item add

The B<add> method takes 2 arguments.  The first is a network or host
specification in "n.n.n.n/n" format.  If the network mask width is
unspecified it is assumed to be 32 - that is, an individual host address
specification.  The second argument is a node value which you would like
to store in the tree (for later retrieval).  For instance a subnet
description might be a useful value.

On success, B<add> returns an empty list.  On failure, i.e. if this
network specification collides with existing subnets, a hash is returned.
The hash's key is in "n.n.n.n/n" format and the corresponding value is
whatever node value was specified when the subnets were B<add>ed to the
tree.

=cut

# add a subnet node to the NetTree object
# e.g.
#
#   $this->add('128.104.16.0/22', { desc => 'DoIT internal - NET staff',
#				    contact => { name => 'hostmaster',
#						 phone => '262-5236' });
#
# arguments:
# arg0 - subnet specification in "dotted.decimal/netmaskwidth" format.
# arg1 - "cargo" - anything you want to store re: this subnet. (Ususally 
#        the cargo would be a hash ref.)
#        The "cargo" argument is optional.  If not specified, it is
#        assumed that the caller is only testing which subnet arg0 would
#        fall within. i.e. That this method is being called only to
#        determine the collisions, the subnet will not be added.
#
# returns:
# on success, an empty list.
# on failure, a hash of collisions.
#    The key to this hash is the subnet specification, and the value is the
#    "cargo" for that subnet.
#
sub add {
   my $self = shift;
   my $spec = shift;
   my $cargo = shift;
   my $ref = $self->{tree};

   my $bits = octets2bits($spec);
   die if ('' eq $bits);

   # be sure that the subnet being added is within this objects' network
   die if (substr($bits, 0, length($self->{bits})) ne $self->{bits});

   my $net = substr($bits, length($self->{bits}));
   my @net = split('', $net);
   my $bit;
   while (-1 != $#net) {
      $bit = shift @net;
      if (defined($ref->[$NetTree::NET])) {
	 return unavailable($self, $ref)
      }
      return () if (!$cargo && !defined($ref->[$bit])); # just checking for subnet - don't add to the tree 
      $ref=\@{$ref->[$bit]}
   }
   if (defined($ref->[0]) || defined($ref->[1])) {
      # collision - spans a smaller subnet
      return (unavailable($self, $ref->[0]),
	      unavailable($self, $ref->[1]))
   }
   if ($cargo) {
      $ref->[$NetTree::CARGO] = $cargo;
      $ref->[$NetTree::SPEC] = $spec;
      $ref->[$NetTree::NET] = $net
   }

   return ()
}

=item prune

The B<prune> method prunes the B<NetTree> object by coallescing adjacent
subnets that have the same node values.  That is, it determines that two
subnets, for instance "10.0.0.0/25" and "10.0.0.128/25", can be combined,
into "10.0.0.0/24", if the node values which were specified when
those subnets were B<add>ed are equivalent.  The node values are compared
with the "eq" operator.  (This should probably be improved so that if the
node values are not references a comparision ("eq") is performed, otherwise
a numeric comparision ("<=>") is performed.  Also, it would be nice to
improve it so that the caller can specify a comparision function such as
with perl's B<sort> function.)

=cut

# prune the NetTree object
# e.g.
#
#    $this->prune;
#
# the prune subroutine prunes branches from the tree if they represent fully
# populated, adjacent subnets.  That is, this function will cause similarly
# defined subnets to be coallesced where it makes sense.
#
sub prune {
   my $self = shift;
   my($ref, $net);
   if (-1 == $#_) {
      $self->{cache} = {}; # purge the cache, it's invalid after pruning
      $ref = $self->{tree}
   } else {
      $ref = shift @_;
      $net = shift @_;
   }

   return '' unless ref($ref);

   if (!defined($ref->[0]) && !defined($ref->[1]) && length($net)) { # leaf node
      die unless defined($ref->[$NetTree::CARGO]);
      return $ref->[$NetTree::CARGO]
   }

   my ($zero, $one);
   if (defined($ref->[0])) {
      $zero = $self->prune($ref->[0], $net . '0')
   }
   if (defined($ref->[1])) {
      $one = $self->prune($ref->[1], $net . '1')
   }

   return '' if (!$zero || !$one);

   if (defined($ref->[0]) && defined($ref->[1]) && $zero eq $one) { # same cargo
      die if $ref->[$NetTree::CARGO];
      $ref->[$NetTree::CARGO] = $zero;
      $ref->[$NetTree::NET] = $net;
      $ref->[$NetTree::SPEC] = bits2octets($self->{bits} . $net);
      undef $ref->[0];
      undef $ref->[1];
      return $ref->[$NetTree::CARGO]
   }
   # else
   return ''
}

=item available

The B<available> method returns an array of network specifications in
"n.n.n.n/n" format.  These "available" subnets are those that have not been
B<add>ed to the given B<NetTree> object.

=cut

# find the available subnets within the network defined by the NetTree object
# e.g.
#     @available = $this->available;
#
# returns:
#    an array of available subnet specifications.
#
sub available {
   my $self = shift;
   my($ref, $net);
   if (-1 == $#_) {
      $ref = $self->{tree}
   } else {
      $ref = shift @_;
      $net = shift @_;
   }
   my @subnets;
   return if $ref->[$NetTree::SPEC]; # unavailable
   # First, handle the case where both $ref->[0] and $ref->[1] are !defined.
   # (this will cause adjacent unused subnets to be coalesced into one larger
   # subnet.)
   if (!defined($ref->[0]) && !defined($ref->[1])) {
      my @octet = bits2octets($self->{bits} . $net . ('0' x (length($self->{bits})-(length($net)))));
      pop(@octet);
      push(@subnets, join('.', @octet) . "/" . (length($self->{bits})+length $net))
   } else {
      if (defined($ref->[0])) {
         push(@subnets, available($self, \@{$ref->[0]}, $net . '0'))
      } else {
         my @octet = bits2octets($self->{bits} . $net . '0' . ('0' x (length($self->{bits})-(length($net)+1))));
         pop(@octet);
         push(@subnets, join('.', @octet) . "/" . (length($self->{bits})+1+length $net))
      }
      if (defined($ref->[1])) {
         push(@subnets, available($self, \@{$ref->[1]}, $net . '1'))
      } else {
         my @octet = bits2octets($self->{bits} . $net . '1' . ('0' x (length($self->{bits})-(length($net)+1))));
         pop(@octet);
         push(@subnets, join('.', @octet) . "/" . (length($self->{bits})+1+length $net))
      }
   }
   return @subnets
}

=item unavailable

The B<unavailable> method returns a hash of "unavailable" subnets - i.e.
those subnets that have been B<add>ed to the given B<NetTree> object.
The hash's key is in "n.n.n.n/n" format and the corresponding value is
whatever node value was specified when that subnet was B<add>ed to the tree.

=cut

# find the unavailable subnets within the network defined by the NetTree object
# e.g.
#     %unavailable = $this->unavailable;
#
# returns:
#    an array of unavailable subnet specifications.
#
sub unavailable {
   my $self = shift;
   my($ref, $net);
   if (-1 == $#_) {
      $ref = $self->{tree}
   } else {
      $ref = shift @_;
      $net = shift @_;
   }

   return () if (!$ref);

   my @subnets;
   if (defined($ref->[0])) {
      push(@subnets, unavailable($self, \@{$ref->[0]}))
   }
   if (defined($ref->[1])) {
      push(@subnets, unavailable($self, \@{$ref->[1]}))
   }
   if ($ref->[$NetTree::CARGO]) {
      push(@subnets,
	   (scalar(bits2octets($self->{bits} . $ref->[$NetTree::NET])) => $ref->[$NetTree::CARGO]))
   }

   return @subnets
}

=item clearcache

The B<clearcache> method forces the cache used by the B<subnets> method
to speed things up.  You might want to do this periodically if your
processes are getting too large.  (Otherwise the cache grows without bound.)

=cut

sub clearcache {
   my $self = shift;

   $self->{cache} = {};
}

=head1 BUGS

Currently only IP networks with 32-bit addresses can be reliably represented
by a B<NetTree> object.

=cut

=head1 AUTHOR

Dave Plonka <plonka@doit.wisc.edu>

Copyright (C) 1998  Dave Plonka.
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

=head1 VERSION

The version number is the module file RCS revision number (B<$Revision: 1.20 $>)
with the minor number printed right justified with leading zeroes to 3
decimal places.  For instance, RCS revision 1.1 would yield a package version
number of 1.001.

This is so that revision 1.10 (which is version 1.010), for example, will
test greater than revision 1.2 (which is version 1.002) when you want to
B<require> a minimum version of this module.

=cut
